class Map # map tools - nenili receno jinak, pocitame v metrech
  @@epsilon=0.01 # zanadbatelné číslo

  # vzdálenost bodů v metrech
  # lineární aproximace distance_bak1 - rozdíl od distance_exact je v ČR menší než desetina promile
  def self.distance(point1, point2) # očekává body jako parametry ve tvaru [lat:Float,lon:Float]
    return nil if point1.nil? or point2.nil?
    begin
      a = point2[1]-point1[1]; b = point2[0]+point1[0]; c = point2[0]-point1[0]
    rescue
      raise [point1,point2].inspect
    end
    6371000 * Math.sqrt(a*a*(0.000523782 + (-0.000005340525438000032 + 0.000000013613111728122937 * b) * b) + 0.000304617 * c * c )
  end

  # Haversinovský vzorec
  def self.distance_exact(point1, point2) # očekává body jako parametry ve tvaru [lat:Float,lon:Float]
    r = 6371000; # v metrech
    fi1 = point1[0]/180 * Math::PI
    fi2 = point2[0]/180 * Math::PI
    dfi = (point2[0]-point1[0])/180 * Math::PI
    dla = (point2[1]-point1[1])/180 * Math::PI
    a = Math.sin(dfi/2) * Math.sin(dfi/2) + Math.cos(fi1) * Math.cos(fi2) * Math.sin(dla/2) * Math.sin(dla/2)
    c = 2 * Math.atan2(Math.sqrt(a), Math.sqrt(1-a))
    d = r * c # v metrech
  end

  # přibližná vzdálenost Pythagorova věta na ekvidistantní válcové projekci
  def self.distance_bak1(point1, point2) # očekává body jako parametry ve tvaru [lat:Float,lon:Float]
    r = 6371000; # v metrech
    x = ((point2[1]-point1[1])/180 * Math::PI) * Math.cos(((point2[0]+point1[0])/180 * Math::PI) /2)
    y = (point2[0]-point1[0])/180 * Math::PI
    d = Math.sqrt(x*x + y*y) * r # v metrech
  end

  # vrati vzdalenost bodu p od primky a,b
  def self.distance_of_point_from_line(p,a,b)
    pa = Map.distance(p,a)
    pb = Map.distance(p,b)
    ab = Map.distance(a,b)
    s = (ab+pa+pb)/2
    plocha = s*(s-pa)*(s-pb)*(s-ab)
    plocha =0 if plocha<0 # ohlida pripady extremne tenkych trojuhelniku jako [49.6217, 16.7839]/[49.7132312, 17.0395032]/[49.7132641, 17.0395952]
    return 2*Math.sqrt(plocha)/ab
  end

  # vrati vzdalenost bodu p od usecky a,b
  def self.distance_of_point_from_segment(p,a,b)
    param = Map.projection_parameter_of_point_to_segment(p,a,b)
    if param>1
      Map.distance(p,b)
    elsif param<0
      Map.distance(p,a)
    else
      Map.distance_of_point_from_line(p,a,b)
    end
  end

  # vrati parametr udavajici polohu projekce bodu p na usecku ab. 0 znamena a, 1 znamena b
  def self.projection_parameter_of_point_to_segment(p,a,b)
    pa = Map.distance(p,a)
    pb = Map.distance(p,b)
    ab = Map.distance(a,b)
    px = Map.distance_of_point_from_line(p,a,b)
    begin
      if (px>pa or px>pb) and (pa>10000 and pb>10000)
        # pokud je p deleko od a,b může vyjít distance od line divně
        return (pa<pb) ? 0 : 1
      else
        ax = Map.safe_sqrt(pa**2-px**2)
        bx = Map.safe_sqrt(pb**2-px**2)
      end
    rescue
      raise "Error with these params p=#{p}, a=#{a}, b=#{b}"
    end
    if (bx>ab) and (ax<bx) # x je na komplementu poloprimky ab
      return -ax/ab
    else
      return ax/ab
    end
  end

  def self.safe_sqrt(value)
    if value<0 
      if value> -@@epsilon
        0
      else
        raise "Error: Odmocnuji zaporne cislo #{value}"
      end
    else
      Math.sqrt(value)
    end
  end
  
  # vrátí nejbližší bod k bodu P na primce AB
  def self.nearest_point_on_line(p,a,b)
    param = Map.projection_parameter_of_point_to_segment(p,a,b)
    return [a[0]+param*(b[0]-a[0]), a[1]+param*(b[1]-a[1])]
  end

  # vrátí nejbližší bod k bodu P na usecce AB
  def self.nearest_point_on_segment(p,a,b)
    param = Map.projection_parameter_of_point_to_segment(p,a,b)
    if param>1
      b
    elsif param<0
      a
    else
      [a[0]+param*(b[0]-a[0]), a[1]+param*(b[1]-a[1])]
    end
  end
  
  def self.to_length_s(length)
    l = length
    if l.abs>1000
      "#{(l/1000).round(1)} km"
    else
      "#{l.round(0)} m"
    end
  end
  
  def self.km_to_length_s(length_in_km); to_length_s(length_in_km*1000) end
  def self.km_to_s(length_in_km); '%.2f km' % length_in_km end
  
  # given points A,B, the result = (B-A)*par + A
  def self.point_on_line_segment_with_parameter(a,b,par)
    [ (b[0]-a[0])*par+a[0], (b[1]-a[1])*par+a[1] ]
  end
  
  def self.genetive_of_barva(barva);  barva.sub(/á$/,'é') end
  def self.rgb_of_barva(barva)
    case barva 
      when 'červená'  then '#c71712'
      when 'modrá'    then '#1761ab'
      when 'zelená'   then '#249140'
      when 'žlutá'    then '#ffd64d'
      when 'černá'    then '#000000'
      when 'oranžová' then '#ffae4d'
      else 'gray'
    end
  end
  
  def self.icon_url_for(barva,typ='pěší')
    case typ.to_s+barva.to_s
      when 'pěšízelená'    then '/assets/TZ-pesi-zelena.svg'
      when 'pěšíčervená'   then '/assets/TZ-pesi-cervena.svg'
      when 'pěšímodrá'     then '/assets/TZ-pesi-modra.svg'
      when 'pěšížlutá'     then '/assets/TZ-pesi-zluta.svg'
      when 'pěšíčerná'     then '/assets/TZ-pesi-cerna.svg'
      when 'místnízelená'   then '/assets/TZ-mistni-zelena.svg'
      when 'místníčervená'  then '/assets/TZ-mistni-cervena.svg'
      when 'místníčerná'    then '/assets/TZ-mistni-cerna.svg'
      when 'místnímodrá'    then '/assets/TZ-mistni-modra.svg'
      when 'místnížlutá'    then '/assets/TZ-mistni-zluta.svg'
      when 'místníoranžová' then '/assets/TZ-mistni-oranzova.svg'
      
      when 'naučnázelená'    then '/assets/TZ-naucna-zelena.svg'
      when 'naučnážlutá'    then '/assets/TZ-naucna-zluta.svg'
      when 'naučnáčervená'  then '/assets/TZ-naucna-cervena.svg'
      when 'naučnámodrá'    then '/assets/TZ-naucna-modra.svg'
      when 'naučnáoranžová' then '/assets/TZ-naucna-oranzova.svg'
      when /(odbočka|chybná|jezdecká|vozíčkářská|lyžařská|zvláštní|cyklistická|neznačená|^)červená/ then '/assets/TZ-otaznik-cervena.png'
      when /(odbočka|chybná|jezdecká|vozíčkářská|lyžařská|zvláštní|cyklistická|neznačená|^)modrá/ then '/assets/TZ-otaznik-modra.png'
      when /(odbočka|chybná|jezdecká|vozíčkářská|lyžařská|zvláštní|cyklistická|neznačená|^)zelená/ then '/assets/TZ-otaznik-zelena.png'
      when /(odbočka|chybná|jezdecká|vozíčkářská|lyžařská|zvláštní|cyklistická|neznačená|^)žlutá/ then '/assets/TZ-otaznik-zluta.png'
      when /(odbočka|chybná|jezdecká|vozíčkářská|lyžařská|zvláštní|cyklistická|neznačená|^)černá/ then '/assets/TZ-otaznik-cerna.png'
      when /(odbočka|chybná|jezdecká|vozíčkářská|lyžařská|zvláštní|cyklistická|neznačená|^)oranžová/ then '/assets/TZ-otaznik-oranzova.png'
      when /^(|chybná|neznačená|odbočka|vozíčkářská)$/ then '/assets/TZ-unknown.png'
      else raise typ.inspect+barva.inspect
    end
  end
  def self.barvy 
    ['červená', 'modrá', 'zelená', 'žlutá']
  end
  
  def self.icon_img_html(barva,typ='pěší',note=nil); "<img title=\"#{barva}/#{typ} #{note}\" class=\"icon\" src=\"#{icon_url_for(barva,typ)}\" />".html_safe end
  
  def self.min_dlat; 111194.85002686501 end
  def self.min_dlon; 69839.37570821225 end

  
  def self.distance_string(point1,point2)
    distance_to_distance_string(Map.distance(point1,point2))
  end
  
  def self.distance_to_distance_string(vzdalenost_v_metrech)
    if vzdalenost_v_metrech.abs>1000
      "#{(vzdalenost_v_metrech/1000).round(1)} km"
    else
      "#{vzdalenost_v_metrech.round(0)} m"
    end
  end
  
end

# Česká republika je ve zóně lat(φ) 48.5 .. 51.1, lon(λ) 12.0 .. 18.9
# Kruh o polomeru 1km se vejde do intervalu  lat +- 0.009 lon +- 0.0136
# 
#  přibližná vzdálenost (http://www.movable-type.co.uk/scripts/latlong.html)
#   var x = (lon2-lon1) * Math.cos((lat1+lat2)/2);
#   var y = (lat2-lat1);
#   var d = sqrt(x*x + y*y) * R;
#
#   derivace cos(x) je -sin(x) v extrémních hodnotách je to 0.748955720 0.7782431485
#   lineární aproximace cos(x) kolem těchto hodnot je 1.3112874971 - 0.0133699977 * x kde x je ve stupních

